/*============================================================================*
 * FILE:                     CEI_VXW_INTERRUPT_FUNCTIONS.C
 *============================================================================*
 *
 *      COPYRIGHT (C) 1998-2019 BY ABACO SYSTEMS, INC. 
 *      ALL RIGHTS RESERVED.
 *
 *      THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *      COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *      THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *      OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *      AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *      SOFTWARE IS HEREBY TRANSFERRED.
 *
 *      THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *      NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS.
 *
 *===========================================================================*
 *
 * FUNCTION:   BusTools/1553-API Library:
 *             Low-level VxWorks/POSIX specific interrupt, thread, and timer functions.
 *
 * EXTERNAL ENTRY POINTS:
 *
 * INTERRUPT AND THREAD FUNCTIONS:
 *
 *   CEI_THREAD_DESTROY    -  Destroys the thread
 *   CEI_EVENT_CREATE      -  Creates an event
 *   CEI_EVENT_DESTROY     -  Destroys the event
 *   CEI_THREAD_CREATE     -  Creates a thread from a pasted function
 *   CEI_WAIT_FOR_EVENT    -  Infinite or timed wait for event
 *   CEI_EVENT_SIGNAL      -  Signals the event
 *   CEI_MUTEX_CREATE      -  Create a mutex
 *   CEI_MUTEX_DESTROY     -  Destroys a mutex
 *
 * SYSTEM SPECIFIC TIME FUNCTIONS:
 *   CEI_GET_TIME          -  Gets the current system time in microseconds
 *   PerformanceCounter    -  High accurate time
 *   MSDELAY               -  Milliseconds Delay function
 *
 *===========================================================================*/

/* $Revision:  8.28 Release $
   Date        Revision
  ----------   ---------------------------------------------------------------
  1/31/2006    Creation date.
  12/23/2019   Modified CEI_THREAD_CREATE and CEI_THREAD_EXIT to support
               VxWorks 7. bch

*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <errnoLib.h>
//#include <vwModNum.h>
#include <sysLib.h>
/*#include <taskLib.h>*/
#ifdef _APEX_
#include <apexLib.h>
#endif 
#include "target_defines.h"
#include "lowlevel.h"
#include "cei_types.h"

struct itimerspec tValue;   /* time to be set */

#ifdef _POSIX_
static struct timespec wdelay;
static pthread_attr_t pt_attr;
static struct sched_param pt_param;
static int pattr_cnt=0;
/***************************************************
*  PTHREAD INTERRUPT FUNCTIONS
***************************************************/

/****************************************************************
*
*  PROCEDURE - CEI_THREAD_DESTROY
*
*  FUNCTION
*     Destroys the thread.
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_THREAD_DESTROY(CEI_THREAD * pThread)
{
   pthread_kill(*pThread, 9);
   if(--pattr_cnt == 0)
      pthread_attr_destroy(&pt_attr);

   return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - CEI_EVENT_CREATE
*
*  FUNCTION
*     Create pthread Event.
*
*  RETURNS
*     BTD_OK
*     BTD_ERR_BADOPEN
****************************************************************/
CEI_INT CEI_EVENT_CREATE(CEI_EVENT * pEvent)
{
   int status;

   status = pthread_cond_init(pEvent, NULL);
   if ( status )
   {
      return BTD_ERR_BADOPEN;
   }
   return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - CEI_EVENT_DESTROY
*
*  FUNCTION
*     Destroys pthread Event.
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_EVENT_DESTROY(CEI_EVENT * pEvent)
{
   pthread_cond_destroy(pEvent);
}

/****************************************************************
*
*  PROCEDURE - CEI_HANDLE_DESTROY
*
*  FUNCTION
*     Destroys Handle.
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_HANDLE_DESTROY(CEI_VOID *pHandle)
{
    return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - p_attr_init
*
*  FUNCTION
*     initializes the pthread_attribute.
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT p_attr_init(int iPriority)
{
   int status;

   status = pthread_attr_init(&pt_attr);
   if(status)
      return BTD_ERR_BADOPEN;

   status = pthread_attr_setschedpolicy(&pt_attr, SCHED_FIFO);
   if(status)
      return BTD_ERR_BADOPEN;

   pt_param.sched_priority = iPriority;

   status = pthread_attr_setschedparam(&pt_attr,&pt_param);
   if(status)
      return BTD_SET_PRIORITY_ERR;

   return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - CEI_THREAD_CREATE
*
*  FUNCTION
*     Create POSIX Thread.
*
*  RETURNS
*     BTD_OK
*     BTD_ERR_BADOPEN
****************************************************************/
CEI_INT CEI_THREAD_CREATE(CEI_THREAD * pThread, CEI_INT iPriority,void * func, void * pFifo)
{
   int status;

   if(pattr_cnt == 0)
   {
      status =  p_attr_init(iPriority);
      if(status)
         return status;
      pattr_cnt++;
   }

   status = pthread_create(pThread, &pt_attr, func, pFifo);

   if ( status )
   {
      return BTD_THRD_CREATE_FAIL;
   }
   return BTD_OK;
}

CEI_INT CEI_THREAD_EXIT(CEI_VOID * status, CEI_VOID * handle)
{
   pthread_exit(0);
}

/****************************************************************
*
*  PROCEDURE - CEI_WAIT_FOR_EVENT
*
*  FUNCTION
*     Waits for specified event.
*
*  RETURNS
*     Nothing
****************************************************************/
CEI_INT CEI_WAIT_FOR_EVENT(CEI_EVENT *pEvent, CEI_INT iDelay, CEI_MUTEX *pMutex)
{
   int retval;

   if(iDelay==INFINITE)
   {
      pthread_mutex_lock(pMutex);
      retval = pthread_cond_wait(pEvent,pMutex);
      pthread_mutex_unlock(pMutex);
      return retval;
   }
   else
   {
      pthread_mutex_lock(pMutex);
      clock_gettime(CLOCK_REALTIME, &wdelay);
      wdelay.tv_nsec = (wdelay.tv_nsec)+(iDelay*1000000);
      retval = pthread_cond_timedwait(pEvent,pMutex,&wdelay);         
      pthread_mutex_unlock(pMutex);
      switch(retval)
      {
         case 0:
            return 0;
         case ETIMEDOUT:
            return BTD_EVENT_WAIT_TIMEOUT;
         default:
            return BTD_EVENT_WAIT_UNKNOWN;
      }         
   }   
}

/****************************************************************
*
*  PROCEDURE - CEI_EVENT_SIGNAL
*
*  FUNCTION
*     Sets the specified event for all waiting threads.
*
*  RETURNS
*     Nothing
****************************************************************/
void CEI_EVENT_SIGNAL(CEI_EVENT *pEvent, CEI_MUTEX *pMutex)
{
   pthread_mutex_lock(pMutex);
   pthread_cond_signal(pEvent);
   pthread_mutex_unlock(pMutex);
}

/****************************************************************
*
*  PROCEDURE - CEI_MUTEX_CREATE
*
*  FUNCTION
*     Creates a pthread mutex.
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_MUTEX_CREATE(CEI_MUTEX *pmutex)
{
   return pthread_mutex_init(pmutex,NULL);
}

/****************************************************************
*
*  PROCEDURE - CEI_MUTEX_LOCK
*
*  FUNCTION
*     Request ownership of a Windows critical section
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_MUTEX_LOCK(CEI_MUTEX *pMutex) {
   pthread_mutex_lock(pMutex);
   return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - CEI_MUTEX_UNLOCK
*
*  FUNCTION
*     Release ownership of a Windows critical section
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_MUTEX_UNLOCK(CEI_MUTEX *pMutex) {
   pthread_mutex_unlock(pMutex);
   return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - CEI_MUTEX_DESTROY
*
*  FUNCTION
*     Destroys a pthread mutex.
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_MUTEX_DESTROY(CEI_MUTEX *pmutex)
{
   pthread_mutex_destroy(pmutex);
}

#else /*_POSIX_*/
#include <objLib.h>  /*  for S_objLib_OBJ_TIMEOUT  */
/***************************************************
*  VxWorks task functions FUNCTIONS
***************************************************/

/****************************************************************
*
*  PROCEDURE - CEI_THREAD_DESTROY
*
*  FUNCTION
*     Destroys the thread.
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_THREAD_DESTROY(CEI_THREAD * pThread)
{
   taskDelete(*pThread);
   
   return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - CEI_EVENT_CREATE
*
*  FUNCTION
*     Create VxWorks Event.
*
*  RETURNS
*     BTD_OK
*     BTD_ERR_BADOPEN
****************************************************************/
CEI_INT CEI_EVENT_CREATE(CEI_EVENT *pEvent)
{
   *pEvent = msgQCreate(64,2,MSG_Q_FIFO);
   if (*pEvent ==  0)
   {
      return BTD_ERR_BADOPEN;
   }
   return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - CEI_EVENT_DESTROY
*
*  FUNCTION
*     Destroys VxWorks Event.
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_EVENT_DESTROY(CEI_EVENT * pEvent)
{
    msgQDelete(*pEvent);
    return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - CEI_HANDLE_DESTROY
*
*  FUNCTION
*     Destroys Handle.
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_HANDLE_DESTROY(CEI_VOID *pHandle)
{
    return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - CEI_THREAD_CREATE
*
*  FUNCTION
*     Create VxWorks Thread.
*
*  RETURNS
*     BTD_OK
*     BTD_ERR_BADOPEN
****************************************************************/
CEI_INT CEI_THREAD_CREATE(CEI_THREAD * pThread, CEI_INT iPriority, void * func, void * iFifo)
{
   char fname[80];
   static int fcount=0;
   CEI_THREAD task_id;
   
   sprintf(fname,"if_%d_%d\n",rand(), fcount++);
   
   task_id = taskSpawn(fname,               /* name of new task (stored at pStackBase) */
                        iPriority,          /* priority of new task */
                        0,                  /* task option word */
                        0x2000,             /* size (bytes) of stack needed plus name */
                        (FUNCPTR)func,      /* entry point of new task */
                        (CEI_ULONG) iFifo,  /* 1st of 10 req'd task args to pass to func */
                        2,3,4,5,6,7,8,9,10);

#ifdef _WRS_CONFIG_LP64
   if(task_id == TASK_ID_ERROR)
#else
   if( task_id == ERROR)
#endif
   {
      return BTD_THRD_CREATE_FAIL;
   }
   
   *pThread = task_id;
   
   return BTD_OK;
}

CEI_INT CEI_THREAD_EXIT(CEI_VOID * status, CEI_VOID * handle)
{
#ifdef _WRS_CONFIG_LP64
   taskDelete((TASK_ID)handle);
#else
   taskDelete((int)handle);
#endif
  
   return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - CEI_WAIT_FOR_EVENT
*
*  FUNCTION
*     Waits for specified event.
*
*  RETURNS
*     Nothing
****************************************************************/
CEI_INT CEI_WAIT_FOR_EVENT(CEI_EVENT *pEvent, CEI_INT iDelay, CEI_MUTEX *pMutex)
{
   ssize_t retval;
   char qmsg[2];

   if(iDelay==INFINITE)
   {
      retval = msgQReceive(*pEvent,qmsg,2,WAIT_FOREVER);
      if(retval == ERROR)
         return -1;
      else
	 return 0;	
   }
   else
   {
      CEI_INT clkRate,tick_time,ticks_to_wait;
	  
      clkRate = sysClkRateGet();
      tick_time = 1000/clkRate;
      ticks_to_wait = (iDelay/tick_time)+1;
      retval = msgQReceive(*pEvent,qmsg,2,ticks_to_wait);

      if(retval == 2)
         return 0;

      if(errnoGet() == S_objLib_OBJ_TIMEOUT )
      {
         return BTD_EVENT_WAIT_TIMEOUT;
      }
      else
      {
         return BTD_EVENT_WAIT_UNKNOWN; 
      }		    
   }
}

/****************************************************************
*
*  PROCEDURE - CEI_EVENT_SIGNAL
*
*  FUNCTION
*     Sets the specified event for all waiting threads.
*
*  RETURNS
*     Nothing
****************************************************************/
CEI_INT CEI_EVENT_SIGNAL(CEI_EVENT *pEvent, CEI_MUTEX *pMutex)
{
   char qmsg[2];
   STATUS status;
  
   status = msgQSend(*pEvent,qmsg,2, NO_WAIT, MSG_PRI_NORMAL);
   if(status == ERROR)
      return BTD_EVENT_SIGNAL_ERR;
   else
      return 0;
}

/****************************************************************
*
*  PROCEDURE - CEI_MUTEX_CREATE
*
*  FUNCTION
*     Creates a VxWorks mutex.
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_MUTEX_CREATE(CEI_MUTEX *pmutex)
{
   *pmutex = semMCreate(SEM_Q_FIFO);
   return 0;
}

/****************************************************************
*
*  PROCEDURE - CEI_MUTEX_LOCK
*
*  FUNCTION
*     Request ownership of a Windows critical section
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_MUTEX_LOCK(CEI_MUTEX *pMutex) {
   semTake(*pMutex,WAIT_FOREVER);
   return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - CEI_MUTEX_UNLOCK
*
*  FUNCTION
*     Release ownership of a Windows critical section
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_MUTEX_UNLOCK(CEI_MUTEX *pMutex) {
  semGive(*pMutex);
  return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - CEI_MUTEX_DESTROY
*
*  FUNCTION
*     Destroys a Windows mutex.
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT CEI_MUTEX_DESTROY(CEI_MUTEX *pmutex)
{
   semDelete(*pmutex);
   return 0;
}
#endif /*_POSIX_*/

/***************************************************************
*  CEI O/S Specific time functions
***************************************************************/

/****************************************************************
*
*  PROCEDURE - CEI_GET_TIME
*
*  FUNCTION
*     Destroys a Windows mutex.
*
*  RETURNS
*     Time in milliseconds
****************************************************************/
unsigned long CEI_GET_TIME(void)
{
   struct timespec ts;        /* Current time, use to get milliseconds */
   
   clock_gettime(CLOCK_REALTIME, &ts);
   return(unsigned long)ts.tv_nsec/1000000l;
}

/* ***************************************************************
*
*  PROCEDURE - PerformanceCounter
*
*  FUNCTION
*     Destroys a Windows mutex.
*
*  RETURNS
*     Time in milliseconds
*************************************************************** */
CEI_UINT64 PerformanceCounter(void)
{
   struct timespec ts;        /* Current time, use to get milliseconds */

   clock_gettime(CLOCK_REALTIME, &ts);
   
   return (CEI_UINT64)((ts.tv_sec*1000000)+ts.tv_nsec/1000);
}

/****************************************************************
*
*  PROCEDURE - MSDELAY
*
*  FUNCTION
*     Common VxWorks Millisecond delay function
*
*  RETURNS
*     nothing
****************************************************************/

void MSDELAY(int msec)  /*  msec from 0 to 999  */
{
#ifdef _APEX_
   RETURN_CODE_TYPE returnCode;
   TIMED_WAIT(1000*msec, &returnCode) ;
#else
   static struct timespec mdelay;

   mdelay.tv_sec = 0;
   mdelay.tv_nsec = msec*1000*1000;

   nanosleep(&mdelay, NULL);
#endif /*  APEX_ */
}
