/*============================================================================*
 * FILE:                    V X W _ I N T _ S E T U P . C
 *============================================================================*
 *
 *      COPYRIGHT (C) 1998-2019 BY ABACO SYSTEMS, INC. 
 *      ALL RIGHTS RESERVED.
 *
 *      THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *      COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *      THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *      OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *      AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *      SOFTWARE IS HEREBY TRANSFERRED.
 *
 *      THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *      NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS.
 *
 *===========================================================================*
 *
 * FUNCTION:   Setting up VxWorks interrupts for PCI interface on x86 and PowerPC.  
 *
 *
 *                  
 *
 * MAJOR INTERNAL ROUTINES:
 *    vbtInterruptClose    Close interrupt threads
 *    vbtSetPolling        Sets up new polling interval
 *    vbtInterruptSetup    Sets up interrupt trheads
 *    intTimerCallback     Perirodic processing function
 *
 *===========================================================================*/

/* $Revision:  8.28 Release $
   Date        Revision
  ----------   ---------------------------------------------------------------
  05/07/2001   Initial Version.
  07/11/2002   Add support for QPMC-1553 v4.54
  01/02/2006   Modified for improve and modular interrupt support
  07/15/2013   This is for the latest include RAR15-XMC-XT and VxWorks653 with APEX
  11/05/2014   Fix BusTools_RT_CbufRead to use vbtReadRelRAM32
  09/03/2019   Renamed devices from gefesDev to avioDev
  12/26/2019   Modified vbtInterruptSetup and int_notice to support VxWorks 7. bch
 */

/*---------------------------------------------------------------------------*
 *                    INCLUDES, DEFINES and GLOBALS
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ioLib.h>
#include <timers.h>
#include <taskLib.h>
#include <errno.h>
#include <errnoLib.h>
#include <signal.h>
#include <version.h>
#include "busapi.h"
#ifdef _APEX_
#include <apexLib.h>
#endif
#include "apiint.h"
#include "globals.h"
#include "btdrv.h"
#ifndef _APEX_
#include "avioVxwDrv.h"
#endif



#define POLLING_TIMER_THREAD_PRIORITY 2

#ifdef DEBUG
#define dprint(a,b) printf(a,b) 
#else
#define dprint(a,b)
#endif

static void (*intFptr[MAX_BTA])(BT_UINT, BT_U32BIT *);
void int_notice(int,int,int,int,int,int,int,int,int,int);
int delay_ms(int ms);

int timerId=0;
static BT_U32BIT bmRecordTime[MAX_BTA] = {0,};
CEI_THREAD polling_timer_thread=0;
CEI_THREAD task_id[MAX_BTA] = {0,0,0,0,
                            0,0,0,0,
                            0,0,0,0,
                            0,0,0,0};
int intUseCount[MAX_BTA] = {0,0,0,0,
                            0,0,0,0,
                            0,0,0,0,
                            0,0,0,0};
int irq[MAX_BTA];
int int_level[MAX_BTA];

/*===========================================================================*
 * ENTRY POINT:           i n t T i m e r C a l l b a c k
 *===========================================================================*
 *
 * FUNCTION:    Windows multimedia timer callback procedure.
 *              See Multimedia function timeSetEvent() for details.
 *
 * DESCRIPTION: Count up the time to the BM Recording event.  Call vbt Notify
 *              for each board which is open, so it can process interrupts.
 *
 *      It will return:
 *              Nothing.
 *===========================================================================*/
 void polling_timer_func(int val1, int val2) 
 {
   BT_INT cardnum=0, rtn=0;
   BT_U32BIT beat=0;

   
   while(1) {
#ifdef _APEX_
       RETURN_CODE_TYPE returnCode;
       TIMED_WAIT(1000*api_polling_interval, &returnCode) ;
#else
     rtn = delay_ms(api_polling_interval);  
     if(rtn)
       continue;
#endif // _APEX_
     for(cardnum=0; cardnum < MAX_BTA; cardnum++) {
       if(bt_inited[cardnum] <= 0)
         continue;
       
       if(board_is_v5_uca[cardnum])
         beat = vbtGetRegister16(cardnum, HWREG_HEART_BEAT);
       else
         beat = vbtGetRegister[cardnum](cardnum, HWREG_HEART_BEATV6);
       if(beat == heartbeat[cardnum])
         channel_status[cardnum].wcs_pulse = 1;
       
       heartbeat[cardnum] = beat;

       /* count the time to next BM recording event */
       bmRecordTime[cardnum] += api_polling_interval;
       vbtNotify[cardnum](cardnum, &bmRecordTime[cardnum]);
     }
   }
 }

 
/*===========================================================================*
 * ENTRY POINT:         V B T I N T E R R U P T S E T U P
 *===========================================================================*
 *
 * FUNCTION:    Setup for interrupt processing.
 *
 * DESCRIPTION: Setup the hardware interrupt if hardware interrupts are
 *              enabled, and setup the polling interrupt if not setup already.
 *
 *      It will return:
 *              Nothing.
 *===========================================================================*/
BT_INT vbtInterruptSetup(BT_UINT cardnum, BT_INT hw_int_enable, BT_INT api_dev)
{
  BT_INT status=0;
  char intproc[16];
  BT_U16BIT * plx_base;
  
  if(cardnum > MAX_BTA)
    return API_BUSTOOLS_BADCARDNUM;

   /*******************************************************************
   *  Setup the hardware interrupt function if interrupts enabled,
   *   but only if HW interrupt is defined and if this is not the
   *   software version of the driver.
   *******************************************************************/
   intFptr[cardnum] = (void *)vbtNotify;

   if(hw_int_enable == API_SW_INTERRUPT) {
     if(timerId == 0) {
       status = vbtSetPolling(api_polling_interval, TIMER_START);
       if(status)
         return status;
     }
     timerId++;

     return API_SUCCESS;
   }
	
   // Clear interrupt by writing a zero to WRITE_INTERRUPT_BIT
   if(board_is_v5_uca[cardnum])
      vbtSetHWRegister(cardnum, HWREG_WRITE_INTERRUPT, 0);
   else
      api_clearhwcbits(cardnum, CR1_HOST_INTERP);

   if(CurrentCardType[cardnum] == QPCI1553) {
      plx_base = (BT_U16BIT *)bt_iobase[cardnum];
      plx_base[38] |= 0x48;
   }

   if(CurrentCardType[cardnum] == QCP1553 || CurrentCardType[cardnum] == QPCX1553) {
      plx_base = (BT_U16BIT *)bt_iobase[cardnum];
      plx_base[52] = flipws(0x0900);
   }

   if(intUseCount[api_dev] == 0) {	
     sprintf(intproc, "interrupt_%d_%d", api_dev, cardnum);
     task_id[api_dev] = taskSpawn (intproc, 64, VXW_TASK_OPTION, 20000, (FUNCPTR)int_notice,cardnum,api_dev,0,0,0,0,0,0,0,0);     
#ifdef _WRS_CONFIG_LP64
     if(task_id[api_dev] == TASK_ID_ERROR)
#else
     if(task_id[api_dev] == ERROR)
#endif
       return BTD_TASK_CREATE;
   }
   intUseCount[api_dev]++;	  
	  
   /* Set CPU_INTERRUPT_ENABLE (bit 14) of 1553_CONTROL_REG */
   if(board_is_v5_uca[cardnum])
     api_writehwreg_or(cardnum, HWREG_CONTROL1, CR1_INT_ENABLE);
   else
     api_sethwcbits(cardnum, CR1_INT_ENABLE);

   return API_SUCCESS;                                                                           
}

/*===========================================================================*
 * ENTRY POINT:         v b t I n t e r r u p t C l o s e
 *===========================================================================*
 *
 * FUNCTION:    close interrupt processing.
 *
 * DESCRIPTION: Close the hardware interrupt if hardware interrupts are
 *              enabled, and disable the polling interrupt if all users are done.
 *
 *      It will return:
 *              Nothing.
 *===========================================================================*/
void vbtInterruptClose(BT_UINT cardnum)
{
   int device=api_device[cardnum];
   
   if(hw_int_enable[cardnum] >= API_HW_INTERRUPT) {
     if((--intUseCount[device]) == 0) {
       /* kills the task for handling hardware interrupts (function "int_notice") */    	 
       taskDelete(task_id[device]);
	 }
   }
   if(hw_int_enable[cardnum] != API_HW_ONLY_INT) {
     if(timerId > 0) {
       if(--timerId == 0) {
         /* kills the task for polling (function "polling_timer_func") */
         taskDelete(polling_timer_thread);
       }
     }
   }
}
 
/*===========================================================================*
 * ENTRY POINT:           v b t S e t P o l l i n g
 *===========================================================================*
 *
 * FUNCTION:    The function sets the polling interval for the application
 *
 * DESCRIPTION: The function stop the timer set the timing interval and
 *              restarts the timer
 *
 *      It will return: status
 *             
 *===========================================================================*/
BT_INT vbtSetPolling(BT_UINT polling, BT_UINT tflag)
{
   BT_INT status=BTD_OK;

   api_polling_interval = polling;
   if(api_polling_interval > MAX_POLLING)
     api_polling_interval = MAX_POLLING;
   if(api_polling_interval < MIN_POLLING)
     api_polling_interval = MIN_POLLING;
   
   if(tflag == TIMER_RESTART) {  
     if(polling_timer_thread != (CEI_THREAD)NULL) {  /* if timer is running stop it */
       CEI_THREAD_DESTROY(&polling_timer_thread);
       polling_timer_thread = (CEI_THREAD)NULL;  /* the timer has been deleted */
     }
     tflag = TIMER_START;
   }
   
   if(tflag == TIMER_START) {
     status = CEI_THREAD_CREATE(&polling_timer_thread, POLLING_TIMER_THREAD_PRIORITY, polling_timer_func, NULL);
     if(status != BTD_OK) {     
       dprint("vbtSetPolling: ERROR - failed create thread %d\n", status);
       status = BTD_TIMER_FAIL;
     }
   }

   return status;
}

/* the ISR for handling hardware interrupts */
void int_notice(int cardnum,int device,int b,int c,int d,int e,int f,int g,int h,int j)
{
   unsigned int intCardnum=0;
   BT_U32BIT rtime=0;
   ssize_t numbytes=0;
   MSG_Q_ID mq_id;
   char sname[80];
  
#if _WRS_VXWORKS_MAJOR >= 6     
   sprintf(sname,"/%s%d", AVIO_MSG_Q_NAME, device);
   mq_id = msgQOpen(sname, 100, 4, MSG_Q_FIFO, 0, NULL);
#ifdef _WRS_CONFIG_LP64
   if(mq_id == MSG_Q_ID_NULL) {
#else
   if(mq_id == NULL) {
#endif
      dprint("int_notice: ERROR - failed to open message queue %s\n", sname);
      return;
   }
#else
   int fd;   
 
   sprintf(sname,"/%s/%d", AVIO_DEV_NAME, device);
   fd = open(sname, O_RDWR, 0);
   if(fd == -1)
   {
	  dprint("int_notice: ERROR - failed to open message queue %s\n", sname);
      return;
   }
   ioctl(fd,PCI_GET_MSG_Q_ID,(BT_U32BIT)&mq_id);
   
   close(fd);
#endif    
   
   while(1) {
	 /* block on event to occur */
     numbytes = msgQReceive(mq_id,(char *)&intCardnum, sizeof(intCardnum), WAIT_FOREVER);
     if((numbytes != ERROR) && (numbytes == sizeof(intCardnum))) {
       cardnum = ((int)(intCardnum)) + (device * 4);
       vbtNotify[cardnum](cardnum, &rtime);
     }
   }
}

#ifndef _APEX_
int delay_ms(int ms) 
{
   struct timespec req, rem;

   req.tv_nsec = (ms * 1000000);
   req.tv_sec  = 0;
   rem.tv_nsec = 0;
   rem.tv_sec  = 0;

   while(1) {
     if(nanosleep(&req, &rem) == 0)
       break;
     if(errno != EINTR)
       return errno;
     req.tv_nsec = rem.tv_nsec;
     req.tv_sec = rem.tv_sec;
   }

   return 0;
}
#endif // _APEX_

