/*============================================================================*
 * FILE:                     M E M _ V X W O R K S . C
 *============================================================================*
 *
 *      COPYRIGHT (C) 1998-2019 BY ABACO SYSTEMS, INC. 
 *      ALL RIGHTS RESERVED.
 *
 *      THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *      COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *      THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *      OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *      AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *      SOFTWARE IS HEREBY TRANSFERRED.
 *
 *      THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *      NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS.
 *
 *===========================================================================*
 *
 * FUNCTION:   BusTools/1553 API VxWorks PCI/ISA Bus interface functions.
 *
 *             This module provides the interface between either BusTools/1553
 *             API and our VxWorks PCI products. 
 *
 *             This module performs the actual I/O and memory-mapping functions
 *             needed to support operation in a VxWorks environment.
 *
 * EXTERNAL ENTRY POINTS:
 *
 *    vbtMapBoardAddresses    Looks up the physical base address of the board
 *                            in the registry then creates a pointer which is
 *                            valid in the current 32-bit environment and may
 *                            be used to directly access the memory-mapped RAM
 *                            on the board.
 *
 *    vbtFreeBoardAddresses   Free any system resources which were obtained in
 *                            the mapping of the physical board address.
 *                       
 *    vbtGetCEIDevInfo        Gets the board information.
 *
 *    vbtMapBoardAddress      Looks up the physical base address of the VME
 *                            board then creates a pointer.
 *
 *    vbtFreeBoardAddress     Free any system resources which were obtained in
 *                            the mapping of the VME board address.
 * VxW_GetVMEConfigAddr
 * vbtMapIoAddress
 * 
 *                        
 * EXTERNAL NON-USER ENTRY POINTS:
 *
 * INTERNAL ROUTINES:
 *
 *===========================================================================*/

/* $Revision:  8.28 Release $
   Date        Revision
  ----------   ---------------------------------------------------------------
  06/04/2001   Configure current API release
  01/03/2002   Integrated CEI-x20 support.skb
  06/07/2002   Added support for CEI-820TX.skb
  11/25/2003   Add support for VME SUP and USR mapping
  07/26/2005   Merged with CEI-x20 version.CEI-x20 v3.93.skb
  10/31/2005   Cleaned up and include target_defines.h
  03/28/2006   Merged with CEI-x20 version.CEI-x20 v3.95.skb
  11/15/2007   Updated CEI-x20 mapping routine.CEI-x20 v4.10.skb
  03/05/2008   Added VxWorks Driver conditional Options
  12/16/2013   Update to support 8.12 API changes
  09/03/2019   Changed gefesDev to avioDev. Changed gefes_ioctl.h to avio_ioctl.h
  12/05/2019   Modified for RTP support. Removed vbtGetPCIConfigRegister, vbtGetIRQ,
               VxW_getCardID, vbtGetPCIDeviceID and vbtFreeBoardAddress. bch 
*/

/*---------------------------------------------------------------------------*
 *                    INCLUDES, DEFINES and GLOBALS
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ioLib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sysLib.h>
#include "lowlevel.h"
#include "cei_types.h"
#if defined(VXW_DRIVER_OPTION) || defined(VXW_VXB_DRIVER)
#include "avioVxwDrv.h"
#endif

#ifdef DEBUG
#define dprint(a,b) printf(a,b) 
#else
#define dprint(a,b)
#endif


#if defined (VXW_DRIVER_OPTION) /* Use Common Driver Mapping */
#ifndef PCI_GET_NUM_CHAN
UINT VxW_GetDevChan(UINT device);
#endif
#ifndef PCI_GET_CSC
UINT VxW_Get1553HIF(int dev);
#endif

/****************************************************************************
*
*  PROCEDURE - vbtMapBoardAddresses
*
*  FUNCTION
*     This routine opens the device and gets the memory-mapped addresses for the 
*     physical base addresses of the board that was created by the IO driver.
*     Will need to call vbtFreeBoardAddress() to have the IO driver free the
*     mapped addresses.
*
*  PARAMETERS
*     See below.
*
*  RETURNS
*     BTD_OK           - mapped address with no error
*     BTD_ERR_NOACCESS - failed to open device
*     SYSTEM_ERROR     - any errors
****************************************************************************/
int vbtMapBoardAddresses(UINT      device, /* Device to open (0 - 9 */
                         PDEVMAP_T pMap)   /* Pointer to region mapping structure */
{
   int fd=0;
   char device_name[80];

   dprint("vbtMapBoardAddresses\n", 0);
   
   sprintf(device_name,"/%s/%d", AVIO_DEV_IO_NAME, device);
   fd = open(device_name, O_RDWR, 0);
   if(fd == -1)
     return BTD_ERR_NOACCESS;
   
   ioctl(fd, PCI_GET_DEV_MAP, (void*)pMap);

   pMap->hKernelDriver = fd;
   
   return BTD_OK;
}

#elif defined(VXW_VXB_DRIVER) /* VxWorks VxBus Option */
/****************************************************************************
*
*  PROCEDURE - vbtMapBoardAddresses
*
*  FUNCTION
*     This routine gets the physical base addresses of the board and 
*     sets the base address pointer(s) for the host.
*
*  PARAMETERS
*     See below.
*
*  RETURNS
*     BTD_OK
****************************************************************************/
int vbtMapBoardAddresses(UINT      device, /* Device to open */
                         PDEVMAP_T pMap)   /* Pointer to region mapping structure */
{
   int i;
   
   dprint("vbtMapBoardAddresses\n", 0);
   
   /* get the bus type */
   switch((Av1553Data->hif[device] & 0x3e) >> 1) {
     case 1:
     case 3:
     case 5:
     case 7:
	   pMap->busType = BUS_PCI;
       break;
     case 9:
     case 11:
     case 12:
     case 14:
     case 15:
     case 17:
     case 18:
	   pMap->busType = BUS_VME;
	   break;
     case 16:
	   pMap->busType = BUS_PCMCIA;
	   break;
     default:
	   pMap->busType = BUS_INTERNAL;
   }

   pMap->device = device;
   
   /* clear the address info */
   pMap->memSections = 0;
   for(i=0; i < 6; i++) {
     pMap->memStartPhy[i] = 0;
     pMap->memLengthBytes[i] = 0;
     pMap->memHostBase[i] = NULL;
   }

   /* if BAR2 is mapped */
   if(Av1553Data->membaseBar2[device] > 0) {
     pMap->memSections = 2;
  	 pMap->flagMapToHost[1] = 1;
     pMap->memHostBase[1] = (char *)Av1553Data->membaseBar2[device];
  	 pMap->memStartPhy[1] = Av1553Data->membaseBar2[device];
     pMap->memLengthBytes[1] = (unsigned int)Av1553Data->size[device];
   }

   /* if BAR0 is mapped */
   if(Av1553Data->membase[device] > 0) {
     if(pMap->memSections == 0) {
       pMap->memSections = 1;
       pMap->memLengthBytes[0] = (unsigned int)Av1553Data->size[device];
     }
     pMap->flagMapToHost[0] = 1;
     pMap->memStartPhy[0] = Av1553Data->membase[device];
	 pMap->memHostBase[0] = (char *)Av1553Data->membase[device];  	 
   }
   
   return BTD_OK;
}
/****************************************************************************************************/
#endif  /* VXW_DRIVER_OPTION */ 

#if defined(VXW_VXB_DRIVER) || defined(VXW_DRIVER_OPTION)
/****************************************************************
*
*  PROCEDURE - vbtFreeBoardAddresses
*
*  FUNCTION
*     This routine frees any system resources which were obtained
*     in the mapping of the physical board address.
*
*  PARAMETERS
*     See below.
*
*  RETURNS
*     BTD_OK
****************************************************************/
CEI_INT vbtFreeBoardAddresses(PDEVMAP_T pMap)
{
   int i;

   dprint("vbtFreeBoardAddresses\n", 0);
   
#if defined (VXW_DRIVER_OPTION)
   /* releases the mapped addresses if this is the last close of the file descriptor */  
   close(pMap->hKernelDriver);
#endif
 
   pMap->memSections = 0;
   for(i=0; i < 6; i++) {
     pMap->flagMapToHost[i] = 0;
     pMap->memStartPhy[i] = 0;
     pMap->memLengthBytes[i] = 0;
     pMap->memHostBase[i] = NULL;
   }
      
   return 0;  
}

/****************************************************************
*
*  PROCEDURE - vbtGetCEIDevInfo
*
*  FUNCTION
*     This routine returns device information about the device.
*
*  RETURNS
*     BTD_ERR_NOWINRT
*     BTD_ERR_BADOPEN
*     BTD_ERR_BADREGISTER
*     BTD_OK
****************************************************************/
int vbtGetCEIDevInfo(UINT device, DEVINFO_T *pInfo)
{
#ifdef VXW_DRIVER_OPTION
   int status=0;
   unsigned int val=0;
   unsigned short cscReg=0;
   char device_name[80];
   int fd=0;

   dprint("vbtGetCEIDevInfo\n", 0);
   
   sprintf(device_name, "/$s/%d", AVIO_DEV_IO_NAME, device);
   fd = open(device_name, O_RDWR, 0);
   if(fd == -1)
     return BTD_ERR_BADOPEN;

   pInfo->VendorID = AVIO_PCI_VENDID;
   /* get the PCI device ID */
   status = ioctl(fd, PCI_GET_DEVICE_ID, &val);
   if(status != 0)
     return BTD_IOCTL_GET_REG;
   pInfo->DeviceID = val & 0xFFFF; 
   
   /* read the CSC register */
#ifndef PCI_GET_CSC
   val = VxW_Get1553HIF(device);
   cscReg = val & 0xFFFF;
#else
   status = ioctl(fd, PCI_GET_CSC, &val);
   if(status != 0)
     return BTD_IOCTL_GET_REG;
   cscReg = val & 0xFFFF;
#endif
   
   /* get the bus type */
   switch((cscReg & 0x3e) >> 1) {
     case 1:
     case 3:
     case 5:
     case 7:
       pInfo->busType = BUS_PCI;
       break;
     case 9:
     case 11:
     case 12:
     case 14:
     case 15:
     case 17:
     case 18:
	   pInfo->busType = BUS_VME;
	   break;
     case 16:
	   pInfo->busType = BUS_PCMCIA;
	   break;
     default:
	   pInfo->busType = BUS_INTERNAL;
   }
   
   /* get the number of channels */
#ifndef PCI_GET_NUM_CHAN
   pInfo->nchan = VxW_GetDevChan(device);
#else
   status = ioctl(fd, PCI_GET_NUM_CHAN, &val);
   if(status != 0)
     return BTD_IOCTL_GET_REG;
   pInfo->nchan = val;
#endif
   
   /* get the number of memory regions */
   status = ioctl(fd, PCI_GET_NUM_REGS, &val);
   if(status != 0)
     return BTD_IOCTL_GET_REG; 
   pInfo->memSections = val;
   
   /* get the base address */
   status = ioctl(fd, PCI_GET_REGION_MEM, &val);
   if(status != 0)
     return BTD_IOCTL_GET_REG; 
   pInfo->base_address = val;

   pInfo->mode = (cscReg >> 11) & 0x1;
   pInfo->irig = (cscReg >> 12) & 0x1;
   
   close(fd);
   
   return 0;
#elif VXW_VXB_DRIVER
   dprint("vbtGetCEIDevInfo\n", 0);
   
   switch((Av1553Data->hif[device] & 0x3e)>>1 )
   {
   case 1:
   case 3:
   case 5:
   case 7:
     pInfo->busType = BUS_PCI;
     break;
   case 9:
   case 11:
   case 12:
   case 14:
   case 15:
   case 17:
   case 18:
	 pInfo->busType = BUS_VME;
	 break;
   case 16:
	 pInfo->busType = BUS_PCMCIA;
	 break;
   default:
	 pInfo->busType = BUS_INTERNAL;
   }

   pInfo->VendorID = AVIO_PCI_VENDID;
   pInfo->DeviceID = Av1553Data->CondorDeviceID[device];
   pInfo->nchan = Av1553Data->dev_chan[device];
   pInfo->mode = (Av1553Data->hif[device] & 0x0800) >> 11;
   pInfo->irig = (Av1553Data->hif[device] & 0x1000) >> 12;
   pInfo->host_interface = Av1553Data->hif[device];
   if(Av1553Data->membaseBar2[device] > 0) {
     pInfo->memSections = 2;
     pInfo->base_address = Av1553Data->membaseBar2[device];
   }
   else {
     pInfo->memSections = 1;
     pInfo->base_address = Av1553Data->membase[device];
   }
   
   return 0;
#endif
}
#endif

#if defined(VXW_VME_PPC) || defined(VXW_VME_X86)
/* NOTE: if a build error occurs that the vme.h file is missing then reference the SBC BSP package for the necessary file(s) to provide the VME address modifiers */ 
#include <vme.h>

#define CARRIER_MAP_A32      0x10   /* Native VME - A32 Addressing        */
#define CARRIER_MAP_A24      0x11   /* Native VME - A24 Addressing        */
#define CARRIER_MAP_A16      0x12   /* Native VME - A16 Addressing        */
#define CARRIER_MAP_A24S     0x13   /* Native VME - A24 VxWorks SUP DATA Addressing */
#define CARRIER_MAP_A32S     0x14   /* Native VME - A32 VxWorks SUP DATA Addressing */
#define CARRIER_MAP_A24SP    0x15   /* Native VME - A24 VxWorks SUP PGM Addressing */
#define CARRIER_MAP_A32SP    0x16   /* Native VME - A32 VxWorks SUP PGM Addressing */
#define CARRIER_MAP_A32UP    0x17   /* Native VME - A32 VxWorks USR PGM Addressing */ 
#define CARRIER_MAP_A24UP    0x18   /* Native VME - A24 VxWorks USR PGM Addressing */ 

static int vme_device_count = 0;
static int vme_device[] = {0,0,0,0,  /* vme_device */
                           0,0,0,0,
                           0,0,0,0,
                           0,0,0,0};

static int vme_device_use[] = {0,0,0,0, /*  vme_device use count */
                               0,0,0,0,
                               0,0,0,0,
                               0,0,0,0};

static int vme_mem_addr[] = {0,0,0,0,
                             0,0,0,0,
                             0,0,0,0, 
                             0,0,0,0};

static int vme_map_addr[] = {0,0,0,0,
                             0,0,0,0,
                             0,0,0,0, 
                             0,0,0,0};


static int vme_io_addr[] = {0,0,0,0,
                             0,0,0,0,
                             0,0,0,0, 
                             0,0,0,0};

int btype,nchan,ci; 
unsigned hif[MAX_BTA];

/*************************************************************************************
*
*  The Following are Utility routine for debugging installations
*     VxW_GetVMEConfigAddr              Returns VME A16 configuration address
*     VxW_GetVMEMemAddr                 Returns VME A32/A24 Memory address
*     VxW_Get_vme_device_count          Returns the number of Condor VME devices
*     VxW_Get_vme_device                Returns the Condor VME
*     VxW_Get_vme_config
*     VxW_GetCardID                     Returns the card type
*     VxW_GetHIF                        Returns the Board Host Interface Register setting
*     VxW_GetDevChan                    Return the number of channels on the device
*     VxW_GetDevIRQ                     Returns the device's IRQ value
*  
**************************************************************************************/
CEI_INT VxW_GetVMEConfigAddr(CEI_UINT cardnum)
{
   return vme_device[cardnum];
}

CEI_INT VxW_GetVMEMemAddr(CEI_INT cardnum)
{
   return vme_mem_addr[cardnum];
}

CEI_INT VxW_Get_vme_device_count(void)
{
   return vme_device_count;
}

CEI_INT VxW_Get_vme_device(CEI_UINT caddr)
{
   int i;

   for(i=0;i<MAX_BTA;i++) {
     if(caddr == vme_device[i])
       return i;
   }
   
   return -1;
}

CEI_INT VxW_Get_vme_config(CEI_INT device)
{
   return vme_device[device];
}

//int vbtGetCEIDevInfo(UINT device, DEVINFO_T *pInfo) {
//	return BTD_NON_SUPPORT;
//}

/*********************************************************************
*   VME Mapping function for 1553 boards
*********************************************************************/

/****************************************************************
*
*  PROCEDURE - vbtMapBoardAddress(base_address,
*                                 map_length, 
*                                 char ** addr)
*
*  PARAMETERS
*     BT_U32BIT  base_address      - Card physical base address.
*     BT_U32BIT  map_length        - Size of region to map in bytes.
*     int        addressing_mode   - VME Addressing Mode.
*     char      ** addr            - Returned address of board.
*     DWORD     junk               - not used
*
*  RETURNS
*     BTD_OK          - Mapped address with no error
****************************************************************/
int vbtMapBoardAddress(unsigned base_address, unsigned map_length, char **addr, void *cardnum, int addressing_mode)
{
   int status;
   int vme_mode;
   int index;

   if(addressing_mode == CARRIER_MAP_A32)
      vme_mode = VME_AM_EXT_USR_DATA;
   else if(addressing_mode == CARRIER_MAP_A32UP)
      vme_mode = VME_AM_EXT_USR_PGM;
   else if(addressing_mode == CARRIER_MAP_A32S)
      vme_mode = VME_AM_EXT_SUP_DATA;
   else if(addressing_mode == CARRIER_MAP_A32SP)
      vme_mode = VME_AM_EXT_SUP_PGM; 
   else if(addressing_mode == CARRIER_MAP_A24)
      vme_mode = VME_AM_STD_USR_DATA;
   else if(addressing_mode == CARRIER_MAP_A24S)
      vme_mode = VME_AM_STD_SUP_DATA;
   else if(addressing_mode == CARRIER_MAP_A24SP)
      vme_mode = VME_AM_STD_SUP_PGM;
   else
      return BTD_ERR_PARAM;

   index = *(int *)cardnum; /* this is the cardnum so base_address can get stored */

   status = sysBusToLocalAdrs(vme_mode, (char *)base_address, addr);
   if(status == OK) {
     vme_map_addr[index] = (unsigned)addr[0];
     vme_mem_addr[index] = base_address;
     return BTD_OK;
   }
   else {
     printf("vbtMapBoardAddress sysBusToLocalAdrs status = %d\n",status);
     return BTD_ERR_BADEXTMEM;
   }
}

/****************************************************************
*
*  PROCEDURE - vbtMapIoAddress(cardnum, base_address, char ** addr)
*
*  VxWorks Version - Flat VME addressing model
*
*  PARAMETERS
*     BT_U32BIT  base_address      - Card I/O base address.
*     char      ** addr            - Returned address of board.
*
*  RETURNS
*     BTD_OK          - Mapped address with no error
****************************************************************/
int vbtMapIoAddress(unsigned base_address, char **addr)
{
   int i;
   static unsigned tcnt=0; 
      
   for(i=0;i<=vme_device_count;i++) {
     if(vme_io_addr[i] == base_address) {
       vme_device[tcnt] = i;
       vme_device_use[i]++;
       tcnt++;
       break;
     }
     else if(vme_io_addr[i] == 0) {
       vme_device[tcnt] = i;
       vme_io_addr[i] = base_address;
       vme_device_use[i]++;
       vme_device_count++;
       tcnt++;
       break;
      }   
   }

   if(sysBusToLocalAdrs(VME_AM_SUP_SHORT_IO, (char *)base_address, addr) != OK)
     return BTD_ERR_BADADDR;
    
  return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - vbtFreeBoardAddresses
*
*  FUNCTION
*     This routine frees any system resources which were obtained
*     in the mapping of the phyical board address.
*
*  PARAMETERS
*     See below.
*
*  RETURNS
*     Tothing
****************************************************************/
CEI_INT vbtFreeBoardAddresses(PDEVMAP_T pMap)
{
   int i;

   for(i=0;i<6;i++)
   {
      if(pMap->flagMapToHost[i])
      {
         pMap->flagMapToHost[i]=0; 
      }
   } 
   return 0;  
}

#endif  /* VXW_VME_PPC and VXW_VME_X86 */  
