/*===================================================================================================================*
 * FILE:                                   A V I O   V X W   D R V . C
 *===================================================================================================================*
 *
 *   COPYRIGHT 2017 - 2019 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED ONLY IN ACCORDANCE WITH THE
 *   TERMS OF SUCH LICENSE AND WITH THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY 
 *   OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE AND SHOULD NOT BE CONSTRUED 
 *   AS A COMMITMENT BY ABACO SYSTEMS, INC.
 *
 *===================================================================================================================*
 *
 * FUNCTION:   Avionics I/O Boards VxBus Gen2 Driver Source
 *
 *===================================================================================================================*
 *
 *   Date    By   Modifications
 * --------  ---  ---------------------------------------------------------------------------------------------------
 * 09/28/17  GK   Created to support the VxBus Gen 2 driver implementation
 *                   Initiation of debug support via AVIO_DEBUG directive no longer supported from the CDF file
 *                   so it was relocated below.  Change #undef AVIO_DEBUG to #define AVIO_DEBUG to print debug
 *                   statements (you must also add the component INCLUDE_KPRINTF in the image kernel configuration).
 * 11/06/17  BDW  Removed the unused structure type PCI_DEVVEND and struct array avioDrvIDList, since they were not
 *                   referenced in the Gen2 driver source.  Renamed local variables geipAvRefCnt and geipPCIDrvNum
 *                   to avioAvRefCnt and avioPciDrvNum respectively for obvious reasons.
 * 09/12/19  BCH  Changed vxbDrvUnlink to avioDrvUnlink. Modified avioInstInit2 to return status and added the call
 *                   to avioDevRegister. Modified avioCfg1553Devices. 
 * 11/07/19  BCH  Added vxbDevMethodRun.                
  *===================================================================================================================*/

#include <errno.h>
#include <semLib.h>
#include <stdio.h>
#include <hwif/buslib/pciDefines.h>
#include <sysLib.h>
#include <vxWorks.h>
#include <version.h>
#include <stdlib.h>
#include <string.h>
#include <hwif/vxBus.h>
#include <hwif/buslib/vxbPciLib.h>
#include <subsys/timer/vxbTimerLib.h>
#include <vxAtomicLib.h>
#include <selectLib.h>
#include <errnoLib.h>
#include <avioVxwDrv.h>
#include <vsbConfig.h>
#include <ctype.h>
#include <sllLib.h>


/*-------------------------------------------------------------------------------------------------------------------*
 * Avionics I/O VxBus Driver Routine Prototypes
 *-------------------------------------------------------------------------------------------------------------------*/
LOCAL STATUS avioInstInit2(VXB_DEV_ID pInst);
LOCAL BOOL avioDevProbe(VXB_DEV_ID pInst);
LOCAL STATUS avioDriverCtrl(VXB_DEV_ID pInst, VXB_DRV_CONTROL *pControl);
LOCAL STATUS avioDevRemove(VXB_DEV_ID pInst, VXB_DRV_CONTROL *pControl);
LOCAL STATUS avioVxbResourceFree(VXB_DEV_ID pInst, AVIO_DEVICE_CTRL_TYPE *pDrvCtrl);

/*-------------------------------------------------------------------------------------------------------------------*
 * External Avionics I/O VxBus Utility Routine Prototypes
 *-------------------------------------------------------------------------------------------------------------------*/
void avioDeviceShow(void);
STATUS vxbDevMethodRun(VXB_METHOD_ID method, void *pArg);

/*-------------------------------------------------------------------------------------------------------------------*
 * Avionics I/O Supported VxBus "Other-Class" Driver Methods and Routines
 *-------------------------------------------------------------------------------------------------------------------*/
LOCAL VXB_DRV_METHOD avioDrvMethods[] =
{
   { VXB_DEVMETHOD_CALL (vxbDevProbe),  avioDevProbe },
   { VXB_DEVMETHOD_CALL (vxbDevAttach), avioInstInit2 },
   { VXB_DEVMETHOD_CALL (driverControl), avioDriverCtrl },
   { VXB_DEVMETHOD_CALL (vxbDrvUnlink),  avioDevRemove },
   VXB_DEVMETHOD_END
};

/*-------------------------------------------------------------------------------------------------------------------*
 * Avionics I/O VxBus Driver Registration Structure
 *-------------------------------------------------------------------------------------------------------------------*/
VXB_DRV avioDrv =
{
   { NULL } ,
   AVIO_DRV_NAME,                    /* Name */
   "Abaco Systems Avionics driver",  /* Description */
   VXB_BUSID_PCI,                    /* Class */
   0,                                /* Flags */
   0,                                /* Reference count */
   avioDrvMethods                    /* Method table */
};


/*-------------------------------------------------------------------------------------------------------------------*
 * Global Data
 *-------------------------------------------------------------------------------------------------------------------*/
/* VXB device identifier */
VXB_DEV_ID avioInst[DRV_MAX_DEVICES] = {NULL};

/* Total number of Avionics I/O Devices Detected - use to determine the Device Number assigned to each board */
LOCAL atomicVal_t avioRefCnt = 0;

/* Driver Instance Counter - every instantiation of the Avionics I/O driver is tracked */
LOCAL atomic_t avioPciDrvNum = ERROR;

/* ISR Message Queue values passed back to the API to indicate which "channel" generated the interrupt */
LOCAL unsigned int mq_data[MAX_CHANNELS] = {0,1,2,3};

#ifdef SPIN_LOCK_PROTECT
/* ISR locking mechanism */
spinlockIsr_t spinlock[DRV_MAX_DEVICES];
#endif 

/* register the driver in the vxBus subsystem */
VXB_DRV_DEF(avioDrv)

/*********************************************************************************************************************
**
** NAME:
** avioInit()
**
** DESCRIPTION:
** Dummy init routine to pull the driver into a kernel image.
**
** RETURN:
** OK
**
**********************************************************************************************************************/
STATUS avioDrvInit ()
{
   return OK;
}

/*********************************************************************************************************************
**
** NAME:
** avioDriverCtrl()
**
** DESCRIPTION:
** Implements the VxBus "other-class driver" driverControl function for the Avionics IO driver.
** This "method" is activated from the device API via execution of the VxBus utility function as follows:
**
**   vxbDevMethodRun (DEVMETHOD_CALL(driverControl), &<controlStructure>)
**
** This function checks the invocation to be the type "get Avionics IO device information", and fills the
** structure if true.
**
** ARGUMENTS:
** pInst    - VxBus device instance
** pControl - method argument
**
** RETURN:
** OK or ERROR
**
**********************************************************************************************************************/
LOCAL STATUS avioDriverCtrl(VXB_DEV_ID pInst, VXB_DRV_CONTROL *pControl)
{
   STATUS status=OK;          /* parameter validation and resource disconnect status */
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl=NULL;   /* reference to the Device Control/Info Structure for this device */
   UINT32 devNum=0;
 
   if((pInst == NULL) || (pControl == NULL) || (pControl->driverName == NULL)) {
     /* Any invocation of this function by the API prior to proper session initialization is an error */
	 status = ERROR;
   }
   else {
     pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)vxbDevSoftcGet(pInst);
     dprint("avioDriverCtrl: invoked for pControl->drvCtrlCmd = %x\n", pControl->drvCtrlCmd);
     dprint("avioDriverCtrl: pInst->pName = %s\n", vxbDevNameGet(pInst));
     dprint("avioDriverCtrl: pControl->driverName = %s\n", pControl->driverName);
     dprint("avioDriverCtrl: pDrvCtrl->dev_id = %x\n", pDrvCtrl->dev_id);

     switch(pControl->drvCtrlCmd) {
       /* the only valid application query of the Avionics VxBus Driver is to map an existing board */
       case VXB_GET_DEV_MAP:
         dprint("avioDriverCtrl: VXB_GET_DEV_MAP request for device %x\n", pDrvCtrl->dev_id);
         devNum = vxbDevUnitGet(pInst);
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->dev_id[devNum] = pDrvCtrl->dev_id;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->irq[devNum] = pDrvCtrl->irq;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->dev_chan[devNum] = pDrvCtrl->dev_chan;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->membase[devNum] = pDrvCtrl->membase;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->membaseBar2[devNum] = pDrvCtrl->membaseBar2;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->baseAddressBar0[devNum] = pDrvCtrl->baseAddressBar0;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->baseAddressBar2[devNum] = pDrvCtrl->baseAddressBar2;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->device[devNum] = pDrvCtrl->device;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->msg_q_id[devNum] = pDrvCtrl->msg_q_id;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->msg_q_com_id[devNum] = pDrvCtrl->msg_q_com_id;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->uca32[devNum] = pDrvCtrl->uca32;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->arinc[devNum] = pDrvCtrl->arinc;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->size[devNum] = pDrvCtrl->size;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->hif[devNum] = pDrvCtrl->hif;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->CondorDeviceID[devNum] = pDrvCtrl->CondorDeviceID;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->device_present[devNum] = 0xcafe0000+devNum;
         ((AVIO_DEVICE_INFO_TYPE *)(pControl->drvCtrlInfo))->number_of_devices = pDrvCtrl->devcnt;
         break;
       default:
         /* Any invocation of this function by the API other than for VXB_GET_DEV_MAP functionality is an error */
         status = ERROR;
     }
   }
   
   return status;
}


/*********************************************************************************************************************
**
** NAME:
** avioDevRemove()
**
** DESCRIPTION:
** Implements the VxBus device removal functionality for the Avionics IO driver.
** This "method" is activated from the device API via execution of the VxBus utility function as follows:
**
**   vxbDevMethodRun (DEVMETHOD_CALL(vxbDrvUnlink), &<controlStructure>)
**
** ARGUMENTS:
** pInst    - VxBus device instance.
** pControl - method argument
**
** RETURN:
** OK or ERROR (check errno)
**
**********************************************************************************************************************/
LOCAL STATUS avioDevRemove(VXB_DEV_ID pInst, VXB_DRV_CONTROL *pControl)
{
   STATUS status=OK;           /* parameter validation and resource disconnect status */
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl=NULL;   /* reference to the Device Control/Info Structure for this device */
   UINT32 devNum=0;
   
  
   if(pInst != NULL) {
     dprint("avioDevRemove: invoked for pControl->drvCtrlCmd = %x\n", pControl->drvCtrlCmd);
     dprint("avioDevRemove: pInst->pName = %s\n", vxbDevNameGet(pInst));
     dprint("avioDevRemove: pControl->driverName = %s\n", pControl->driverName);

     /* if this instance is a defined session, check if it is active */
     pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)vxbDevSoftcGet(pInst);
     if(pDrvCtrl != NULL) {
       dprint("avioDevRemove: device %d\n", (int)pDrvCtrl->device);
       devNum = vxbDevUnitGet(pInst);

       /* Disconnect the respective ISR from this session, note the handler routine is specific to the board type. */
       switch(pDrvCtrl->CondorDeviceID) {
         /* 1553 PCI/PMC/XMC products */
         case AVIO_PCI_QPM_DEVID:
         case AVIO_PCI_QCP_DEVID:
         case AVIO_PCI_XMC_DEVID:
         case AVIO_PCI_XMC2_DEVID:
         case AVIO_PCI_R15PMC_DEVID:
         case AVIO_PCI_R15_MPCIE_DEVID:
           /* v6 firmware and later (UCA32) */
           if(pDrvCtrl->uca32) {
             vxbIntDisable(pInst, pDrvCtrl->pIntRes);
             vxbIntDisconnect(pInst, pDrvCtrl->pIntRes);
           }
           else {  /* v5 firmware (UCA16) */
             vxbIntDisable(pInst, pDrvCtrl->pIntRes);
             vxbIntDisconnect(pInst, pDrvCtrl->pIntRes);
           }
           avioVxbResourceFree(pInst, pDrvCtrl);
           status = msgQDelete(pDrvCtrl->msg_q_id);
           break;
               
         /* 1553/ARINC combo cards are UCA32-based, with two message queues to remove */
         case AVIO_PCI_RAR15_DEVID:
         case AVIO_PCI_RAR15XT_DEVID:
           vxbIntDisable(pInst, pDrvCtrl->pIntRes);
           vxbIntDisconnect(pInst, pDrvCtrl->pIntRes);
           avioVxbResourceFree(pInst, pDrvCtrl);
           status = msgQDelete(pDrvCtrl->msg_q_id);
           status = msgQDelete(pDrvCtrl->msg_q_com_id);
         /* commercial PCI/PMC products with a PLX PCI9056 interface */
         case AVIO_PCI_CEI830_DEVID:
         case AVIO_PCI_R830RX_DEVID:
         case AVIO_PCI_R830X820_DEVID:
         case AVIO_PCI_RCEI830A_DEVID:
         case AVIO_PCI_RAR_CPCI_DEVID:
         case AVIO_PCI_PSER_DEVID:
           vxbIntDisable(pInst, pDrvCtrl->pIntRes);
           vxbIntDisconnect(pInst, pDrvCtrl->pIntRes);
           avioVxbResourceFree(pInst, pDrvCtrl);
           status = msgQDelete(pDrvCtrl->msg_q_com_id);
           break;
               
         case AVIO_PCI_RARXMC_DEVID:        /* commercial PCI Express products */
         case AVIO_PCI_RAR_MPCIE_DEVID:
           vxbIntDisable(pInst, pDrvCtrl->pIntRes);
           vxbIntDisconnect(pInst, pDrvCtrl->pIntRes);
           avioVxbResourceFree(pInst, pDrvCtrl);
           status = msgQDelete(pDrvCtrl->msg_q_com_id);
           break;

         default:
           status = ERROR;
           break;
       }

       /* proceed only if the remove request was for a valid board */
       if(status == OK) {
         /* Check to see if this is the last instance of this driver, if so de-register it from the o/s */
         if(vxAtomicDec(&avioPciDrvNum) == ERROR) {
           /* last instance of this driver */
           /* free dynamically allocated device control structure */
           if(pDrvCtrl->name) {
             free(pDrvCtrl->name);
           }
           free(pDrvCtrl);
           vxbDevSoftcSet(pInst, NULL);
           avioInst[devNum] = NULL;
         }
       }
    } 
    else {
       /* invalid session referenced - pDrvCtrl is NULL */
       errnoSet(ENXIO);
       status = ERROR;
    }
  }
  else {
    /* bad argument - pInst is NULL */
    errnoSet(EINVAL);
    status = ERROR;
  }
   
  return status;
}

/*********************************************************************************************************************
**
** NAME:
** avioDevProbe()
**
** DESCRIPTION:
** Queries each VxBus device instance for presence of a Avionics PCI/PCIe IO device.
**
** ARGUMENTS:
** pInst - VxBus device instance.
**
** RETURN:
** OK or ERROR (check errno)
**
**********************************************************************************************************************/
LOCAL STATUS avioDevProbe(VXB_DEV_ID pInst)
{
   STATUS status=OK; /* parameter and device reference validation status */
   VXB_DEV_ID pParentBus=NULL;
   VXB_BUSTYPE_ID pBusType=NULL;


   dprint("avioDevProbe: device\n", 0);
   
   if(pInst != NULL)
     pParentBus = vxbDevParent(pInst);
   if(pParentBus != NULL)
     pBusType = vxbDevClassGet(pParentBus);

   /* validate the parameter and pertinent information therein */
   if((pInst == NULL) || (pParentBus == NULL) || (pBusType == NULL)) {
     errnoSet(EINVAL);
     status = ERROR;
   }
   else {
     /* Only PCI devices are valid Avionics VxBus devices, for today */
     if(VXB_BUSID_MATCH(pBusType, VXB_BUSID_PCI)) {
       UINT32 subId=0;
      
       /* read this device's subsystem device and vendor ID from PCI configuration space */
       VXB_PCI_BUS_CFG_READ(pInst, PCI_CFG_VENDOR_ID, 4, subId);

       switch(subId) {
         case AVIO_PMC_ID:
         case AVIO_QCP_ID:
         case AVIO_XMC_ID:
         case AVIO_XMC2_ID:
         case AVIO_RAR15_ID:
         case AVIO_RAR15XT_ID:
         case AVIO_R15PMC_ID:
         case AVIO_PCI_R15_MPCIE_ID:
         case AVIO_RAR_CPCI_ID:
         case AVIO_RP708_ID:
         case AVIO_P708_ID:
         case AVIO_PSER_ID:
         case AVIO_PDIS_ID:
         case AVIO_CEI830_ID:
         case AVIO_R830RX_ID:
         case AVIO_R830X820_ID:
         case AVIO_RCEI830A_ID:
         case AVIO_RARXMC_ID:
         case AVIO_PCI_RAR_MPCIE_ID:
           dprint("avioDevProbe:  detected valid board - PCI Device/Vendor ID = 0x%X\n", subId);
           break;
         default:
           status = ERROR;
           break;
       }
     }
     else {
       /* device not found */
       status = ERROR;
     }
   }

   return status;
}

/*********************************************************************************************************************
**
** NAME:
** avioDeviceShow()
**
** DESCRIPTION:
** Displays information for each Avionics PCI/PCIe IO device discovered and supported by this VxBus driver.
**
** ARGUMENTS:
** N/A
**
** RETURN:
** NA but check errno (EINVAL)
**
**********************************************************************************************************************/
void avioDeviceShow(void)
{
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl=NULL;   /* reference to the Device Control/Info Structure for this device */
   int devnum=0;                           /* device index for displaying information for all discovered devices */

  
   if(avioInst == NULL) {
     printf("\nNo Avionics I/O Devices Detected\n");
     errnoSet(EINVAL);
     return;
   }

   printf("Avionics I/O Device(s) Detected:\n\n");
#ifdef _WRS_CONFIG_LP64
   printf("Unit  Device ID   Base Address        IRQ   Channels  UCA32  ARINC\n");
#else
   printf("Unit  Device ID   Base Address  IRQ   Channels  UCA32    ARINC\n");
#endif

   printf("-----------------------------------------------------------------------------\n");
   while (avioInst[devnum]) {
     pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *) vxbDevSoftcGet(avioInst[devnum]);
#ifdef _WRS_CONFIG_LP64
     printf("%2d    %4x        %p  %d        %d        %d      %d  \n",
#else
     printf("%2d    %4x        0x%08X    0x%X     %2d       %d        %d  \n",
#endif
     pDrvCtrl->device,
     pDrvCtrl->CondorDeviceID,
     pDrvCtrl->membase,
     pDrvCtrl->irq,
     pDrvCtrl->dev_chan,
     pDrvCtrl->uca32,
     pDrvCtrl->arinc);
     devnum++;
   }
}


/*********************************************************************************************************************
** NAME:
** avioCfg1553Devices()
**
** DESCRIPTION:
** This routine configures Avionics 1553 device information in the structure available to the application.
**
** ARGUMENTS:
** AVIO_DEVICE_CTRL_TYPE * pDrvCtrl    device control & info structure
**
** RETURNS:
** OK or ERROR
**********************************************************************************************************************/
int avioCfg1553Devices(AVIO_DEVICE_CTRL_TYPE * pDrvCtrl)
{
   unsigned short hif=0, dev_id=0;
   unsigned short *base_addr=NULL;

   if(pDrvCtrl->region_count == 2)
     base_addr = (unsigned short *)pDrvCtrl->baseAddressBar2;
   else
     base_addr = (unsigned short *)pDrvCtrl->baseAddressBar0;

   if(base_addr != NULL) {
      /* read the board's "CSC" register for configuration information */
      hif = *base_addr;
      if((hif == 0xFFFF) || (hif == 0)) {
        dprint("avioCfg1553Devices: ERROR - invalid CSC\n", 0);
        return ERROR;
      }
      flipw(&hif);

      pDrvCtrl->hif = hif;
      dev_id = (unsigned short)((hif & 0x3e) >> 1);
      pDrvCtrl->dev_chan = (hif & 0x7c0) >> 6;
      pDrvCtrl->uca32 = (hif & 0x4000) ? 1 : 0;
      pDrvCtrl->arinc = ((dev_id == AVIO_MUL_PROT_TYP2) || (dev_id == AVIO_MUL_PROT_TYP1)) ? 1 : 0;
   }

   dprint("avioCfg1553Devices:  PCI Device ID = %x\n", pDrvCtrl->CondorDeviceID);
   dprint("avioCfg1553Devices:  base addr = 0x%p\n", pDrvCtrl->membase);
   dprint("avioCfg1553Devices:  CSC = 0x%x\n", pDrvCtrl->hif);
   dprint("avioCfg1553Devices:  dev id = 0x%x\n", pDrvCtrl->dev_id);
   dprint("avioCfg1553Devices:  uca32 = %d\n", pDrvCtrl->uca32);
   dprint("avioCfg1553Devices:  ARINC = %d\n", pDrvCtrl->arinc);
   dprint("avioCfg1553Devices:  chan = %d\n", pDrvCtrl->dev_chan);
   dprint("avioCfg1553Devices: Complete for device %d\n", pDrvCtrl->devcnt);

   return OK;
}


/*********************************************************************************************************************
** NAME:
** avioConfigDevices()
**
** DESCRIPTION:
** This routine configures Avionics generic device information in the structure available to the application,
** as well as handling registration of the device BAR memory region(s) to VxWorks.
**
** ARGUMENTS:
** VXB_DEV_ID pInst -  Avionics PCI/PCIe IO device driver instance
**
** RETURNS:
** N/A but check errno
**
**********************************************************************************************************************/
void avioConfigDevices(VXB_DEV_ID pInst)
{
   UINT16 devID=0;                          /* PCI Device Identification value */
   unsigned int commercialHostInterface=0;  /* commercial (ARINC/P-SER) device Global Enable register value */
   STATUS driverConfigStatus=OK;            /* validity of the device referenced by the open session */
   VXB_RESOURCE *pMemRes=NULL;
   VXB_RESOURCE_ADR *pResAdr=NULL;
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl=NULL;    /* reference to the Device Control/Info Structure for this device */


   pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)vxbDevSoftcGet(pInst);

   /* the only valid Avionics device configuration is one where the first mapped region is a memory region (not I/O) */
   if(VXB_RES_TYPE(pDrvCtrl->pMemRes0->id) == VXB_RES_MEMORY) {
      /* Initialize some generic information regarding this device */
      pDrvCtrl->bustype = 2;
      pDrvCtrl->mem_base_reg = 0;
      pDrvCtrl->region_count = 1;
      pDrvCtrl->size = BAR_SIZE_8MB;
      pDrvCtrl->dev_chan = 0;
      pDrvCtrl->uca32 = 0;
      VXB_PCI_BUS_CFG_READ(pInst, PCI_CFG_DEVICE_ID, 2, devID);
      devID &= 0xFFFF;
      pDrvCtrl->CondorDeviceID = (UINT16)devID;
      dprint("avioConfigDevices: Device ID = %x\n",devID);

      /* for 1553 devices, the dev_id structure member is translated into some sort of board type, not so for commercial */
      switch (devID) {
         case AVIO_PCI_QPM_DEVID:
            pDrvCtrl->dev_id = DRV_QPM1553;
            break;
         case AVIO_PCI_QCP_DEVID:
            pDrvCtrl->dev_id = DRV_QCP1553;
            break;
         case AVIO_PCI_XMC_DEVID:
            pDrvCtrl->dev_id = DRV_RXMC1553;
            break;
         case AVIO_PCI_R15PMC_DEVID:
            pDrvCtrl->dev_id = DRV_R15PMC;
            break;
         case AVIO_PCI_XMC2_DEVID:
            pDrvCtrl->dev_id = DRV_R15XMC2;
            break;
         case AVIO_PCI_RAR15_DEVID:
            pDrvCtrl->dev_id = DRV_RAR15XMC;
            break;
         case AVIO_PCI_RAR15XT_DEVID:
            pDrvCtrl->dev_id = DRV_RAR15XMCXT;
            pDrvCtrl->size = BAR_SIZE_16MB;
            break;
         case AVIO_PCI_R15_MPCIE_DEVID:
            pDrvCtrl->dev_id = DRV_R15_MPCIE;
            break;

         /* PCIbus Commercial Avionics Devices (PLX based with host ifc in BAR2) */
         case AVIO_PCI_CEI830_DEVID:
         case AVIO_PCI_R830RX_DEVID:
         case AVIO_PCI_R830X820_DEVID:
         case AVIO_PCI_RCEI830A_DEVID:
         case AVIO_PCI_RAR_CPCI_DEVID:
         case AVIO_PCI_RP708_DEVID:
         case AVIO_PCI_P708_DEVID:
         case AVIO_PCI_PSER_DEVID:
         case AVIO_PCI_PDIS_DEVID:
            pDrvCtrl->mem_base_reg = 2;
            pDrvCtrl->region_count = 2;
            pDrvCtrl->dev_id = devID;
            if ((devID == AVIO_PCI_RP708_DEVID) || (devID == AVIO_PCI_P708_DEVID) || (devID == AVIO_PCI_PSER_DEVID) || (devID == AVIO_PCI_PDIS_DEVID)) {
               pDrvCtrl->size = BAR_SIZE_4MB;
            }
            else {
               pDrvCtrl->size = BAR_SIZE_512KB;
            }
            /* Allocate and map vxbus memory resources for BAR2 */
            pMemRes = vxbResourceAlloc (pInst, VXB_RES_MEMORY, 2);
            if (pMemRes == NULL) {
               dprint("avioConfigDevices: ERROR - can't allocate pMemRes\n", 0);
            }
            else {
               pResAdr = (VXB_RESOURCE_ADR *)pMemRes->pRes;
               if (pResAdr == NULL) {
                  dprint("avioConfigDevices: ERROR - invalid resource address\n",0);
                  vxbResourceFree (pInst, pMemRes);
                  pDrvCtrl->pMemRes2 = NULL;
                  pDrvCtrl->baseAddressBar2 = NULL;
                  pDrvCtrl->membaseBar2 = 0UL;
                  pDrvCtrl->baseHandle[2] = NULL;
               }
               else {
                  pDrvCtrl->pMemRes2 = pMemRes;
                  pDrvCtrl->baseAddressBar2 = (void *) pResAdr->virtAddr;
                  pDrvCtrl->membaseBar2 = (unsigned long) pResAdr->virtAddr;
                  pDrvCtrl->baseHandle[2] = (void *) pResAdr->start;
               }
            }
            dprint("avioConfigDevices: Device BAR2 = %x\n", pDrvCtrl->baseHandle[2]);
            dprint("avioConfigDevices: Device Virtual BAR2 = %x\n", pDrvCtrl->membaseBar2);
#ifdef VXB_VER_5_0_0
            pInst->pRegBasePhys[2] = pInst->pRegBase[2];
#endif
            commercialHostInterface = *(unsigned int *)pDrvCtrl->baseAddressBar2;
            pDrvCtrl->hif = (unsigned short)commercialHostInterface;
            pDrvCtrl->arinc = 1;
            break;
         /* PCIe Commercial Avionics Devices */
         case AVIO_PCI_RARXMC_DEVID:
         case AVIO_PCI_RAR_MPCIE_DEVID:
            pDrvCtrl->dev_id = devID;
            pDrvCtrl->size = BAR_SIZE_512KB;
            pDrvCtrl->arinc = 1;
            commercialHostInterface = *(unsigned int *)pDrvCtrl->baseAddressBar0;
            pDrvCtrl->hif = (unsigned short)commercialHostInterface;
            break;
         default:
            dprint("avioConfigDevices:  Unknown device ID %x\n", devID);
            driverConfigStatus = ERROR;
            break;
      }

      if (driverConfigStatus == OK) {
         pDrvCtrl->device = vxbDevUnitGet (pInst);
         dprint("avioConfigDevices:  Base address = %p\n", pDrvCtrl->baseAddressBar0);
         dprint("avioConfigDevices:  Complete with unit %d\n", pDrvCtrl->device);
      }
      else {  /* error */
         errnoSet(EFAULT);
      }
   }
   else {  /* error */
      vxAtomicDec(&avioRefCnt);
      errnoSet(EFAULT);
   }
}

/*********************************************************************************************************************
** NAME:
** avioInstInit2()
**
** DESCRIPTION:
** This is the main Avionics IO devices VxBus driver post-kernel initialization routine.
**
** ARGUMENTS:
** VXB_DEV_ID pInst
**
** RETURNS
**    OK, ERROR (check errno)
***********************************************************************************************************************/
LOCAL STATUS avioInstInit2(VXB_DEV_ID pInst)
{
   UINT32 deviceIndex=0;                   /* device number referenced in the call */
   char mq_name[40];                       /* message queue name */
   char mq2_name[40];                      /* message queue name */
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl=NULL;   /* reference to the Device Control/Info Structure for this device */
   VXB_RESOURCE *pIntRes=NULL;             /* VxBus2 interrupt resource */
   VXB_RESOURCE *pMemRes=NULL;
   VXB_RESOURCE_ADR * pResAdr=NULL;
   VXB_RESOURCE_IRQ * pResIrq=NULL;

   dprint("avioInstInit2: begin execution\n", 0);
   
   if(!pInst) {  /* bad argument - pInst is NULL */
     errnoSet(EINVAL);
     dprint("avioInstInit2: Instance pInst is NULL\n", 0);
     return ERROR;
   }

   /* Increment the reference counter indicating the number of instances this driver has been initialized. */
   deviceIndex = (UINT32)vxAtomicInc(&avioRefCnt);
   vxbDevUnitSet (pInst, deviceIndex);

   /* create a local copy of this device driver instance */
   avioInst[deviceIndex] = pInst;

   pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)malloc(sizeof(AVIO_DEVICE_CTRL_TYPE));

   while(pDrvCtrl) {
     int i=0;

     /* init device data structure */
     memset((void *)pDrvCtrl, 0, sizeof(AVIO_DEVICE_CTRL_TYPE));

     /* Set these next items to a reasonable default.  If there is more than
      * 1 BAR, the config routine that gets called later will detect that
      * and set these values appropriately.
      */
     pDrvCtrl->mem_base_reg = 1;
     pDrvCtrl->region_count = 1;

     /* create device name */
     if((i = snprintf(NULL, 0, "%s/%d", AVIO_DEV_NAME, deviceIndex)) > 0) {
       if((pDrvCtrl->name = malloc(i + 1)) > 0) {
         if((i = snprintf(pDrvCtrl->name, i + 1, "%s/%d", AVIO_DEV_NAME, deviceIndex)) > 0) {
           /* increment the number of driver instances / open sessions */
           vxAtomicInc(&avioPciDrvNum);

           /* connect to VxBus */
           vxbDevSoftcSet(pInst, pDrvCtrl);
           pDrvCtrl->vxbDevId = pInst;
           dprint("avioInstInit2: avioPciDrvNum = %d\n", avioPciDrvNum);
           dprint("avioInstInit2: calling avioConfigDevices for device %s\n", pDrvCtrl->name);

           /* Get VxBus memory resource for BAR0.  Defer BAR2 to later. */
           pMemRes = vxbResourceAlloc(pInst, VXB_RES_MEMORY, 0);
           if(pMemRes == NULL) {
             dprint("avioInstInit2: ERROR - can't allocate memory resources\n", 0);
             continue;
           }
           else {
             pResAdr = (VXB_RESOURCE_ADR *)pMemRes->pRes;
             if(pResAdr == NULL) {
               dprint("avioInstInit2: ERROR - invalid resource address\n",0);
               vxbResourceFree(pInst, pMemRes);
               pDrvCtrl->pMemRes0 = NULL;
               pDrvCtrl->baseAddressBar0 = NULL;
               pDrvCtrl->membase = 0UL;
               pDrvCtrl->baseHandle[0] = NULL;
               continue;
             }
             else {
               pDrvCtrl->pMemRes0 = pMemRes;
               pDrvCtrl->baseAddressBar0 = (void *) pResAdr->virtAddr;
               pDrvCtrl->size = pResAdr->size;
               pDrvCtrl->membase = (unsigned long) pResAdr->virtAddr;
               pDrvCtrl->baseHandle[0] = (void *) pResAdr->start;
             }
           }

           /* Get VxBus Interrupt Resource */
           pIntRes = vxbResourceAlloc(pInst, VXB_RES_IRQ, 0);
           if(pIntRes == NULL) {
             dprint("avioInstInit2: ERROR - can't allocate interrupt resource %d\n", avioPciDrvNum);
           }
           pDrvCtrl->pIntRes = pIntRes;

           /* define the contents of the device control & information structure */
           /* vxbus memory resource for BAR2 will be setup in this function, if needed */
           avioConfigDevices(pInst);

           /* connect and enable the interrupt. */
           switch(pDrvCtrl->CondorDeviceID) {
             case AVIO_PCI_QPM_DEVID:
             case AVIO_PCI_QCP_DEVID:
             case AVIO_PCI_XMC_DEVID:
             case AVIO_PCI_XMC2_DEVID:
             case AVIO_PCI_R15PMC_DEVID:
             case AVIO_PCI_R15_MPCIE_DEVID:
             /* create a 1553 Message Queue for this device for interrupt notification purposes */
             sprintf(mq_name,"/%s%d", AVIO_MSG_Q_NAME, deviceIndex);
             pDrvCtrl->msg_q_id = msgQOpen(mq_name, 100, 4, MSG_Q_FIFO, OM_CREATE, NULL);
             if(pDrvCtrl->msg_q_id == MSG_Q_ID_NULL) {
               dprint("avioInstInit2: ERROR - can't create message queue %s\n", mq_name);
             }
                         	 
			 dprint("avioInstInit2: calling avioCfg1553Devices for device %d\n", avioPciDrvNum);
             /* further initialize information for all 1553 devices */
             avioCfg1553Devices(pDrvCtrl);

             if(pDrvCtrl->uca32) {
               dprint("avioInstInit2: setting up UCA32 interrupts\n", 0);
               if(vxbIntConnect(pInst, pIntRes, (VOIDFUNCPTR)intProcess32, pInst) != OK) {
                 dprint("avioInstInit2: ERROR - can't connect ISR\n", 0);
               }
               else {
#ifdef SPIN_LOCK_PROTECT
                 spinLockIsrInit(&spinlock[deviceIndex],0);
#endif
                 if(vxbIntEnable(pInst, pIntRes) != OK) {
                   dprint("avioInstInit2: ERROR - can't enable interrupt\n", 0);
                 }
               }
             }
             else {
               dprint("avioInstInit2: setting up UCA interrupts\n", 0);
               if(vxbIntConnect(pInst, pIntRes, (VOIDFUNCPTR)intProcess, pInst) != OK) {
                 dprint("avioInstInit2: ERROR - can't connect ISR\n", 0);
               }
               else {
#ifdef SPIN_LOCK_PROTECT
                 spinLockIsrInit(&spinlock[deviceIndex],0);
#endif
                 if(vxbIntEnable(pInst, pIntRes) != OK) {
                   dprint("avioInstInit2: ERROR - can't enable interrupt\n", 0);
                 }
               }
             }
             break;
             case AVIO_PCI_RAR15_DEVID:
             case AVIO_PCI_RAR15XT_DEVID:
               /* create a 1553 Message Queue for this device for interrupt notification purposes */
               sprintf(mq_name, "/%s%d", AVIO_MSG_Q_NAME, deviceIndex);
               pDrvCtrl->msg_q_id = msgQOpen(mq_name, 100, 4, MSG_Q_FIFO, OM_CREATE, NULL);
               if(pDrvCtrl->msg_q_id == MSG_Q_ID_NULL) {
                 dprint("avioInstInit2: ERROR - can't create message queue %s\n", mq_name);
               }
               
               /* create a Commercial Message Queue for this device for interrupt notification purposes */
               sprintf(mq2_name, "/%s%d", AVIO_COM_MQ_NAME, deviceIndex);
               pDrvCtrl->msg_q_com_id = msgQOpen(mq2_name, 100, 4, MSG_Q_FIFO, OM_CREATE, NULL);
               if(pDrvCtrl->msg_q_com_id == MSG_Q_ID_NULL) {
                 dprint("avioInstInit2: ERROR - can't create message queue %s\n", mq2_name);
               }
               
               dprint("avioInstInit2: calling avioCfg1553Devices\n", 0);
               /* further initialize information for all 1553 devices */
               avioCfg1553Devices(pDrvCtrl);

               dprint("avioInstInit2: setting up UCA32 interrupts\n", 0);
               if(vxbIntConnect(pInst, pIntRes, (VOIDFUNCPTR)intProcess32, pInst) != OK) {
                 dprint("avioInstInit2: ERROR - can't connect ISR\n", 0);
               }
               else {
#ifdef SPIN_LOCK_PROTECT
                 spinLockIsrInit (&spinlock[deviceIndex],0);
#endif
                 if(vxbIntEnable(pInst, pIntRes) != OK) {
                   dprint("avioInstInit2: ERROR - can't enable interrupt\n", 0);
                 }
               }
               break;
               /* Setup ISR for PLX-based PMC devices (x30 & P-SER) */
               case AVIO_PCI_CEI830_DEVID:
               case AVIO_PCI_R830RX_DEVID:
               case AVIO_PCI_R830X820_DEVID:
               case AVIO_PCI_RCEI830A_DEVID:
               case AVIO_PCI_RAR_CPCI_DEVID:
               case AVIO_PCI_PSER_DEVID:
                 /* create a Commercial Product Message Queue for this device for interrupt notification purposes */
                 sprintf(mq_name,"/%s%d", AVIO_COM_MQ_NAME, deviceIndex);
                 pDrvCtrl->msg_q_com_id = msgQOpen(mq_name, 100, 4, MSG_Q_FIFO, OM_CREATE, NULL);
                 if(pDrvCtrl->msg_q_com_id == MSG_Q_ID_NULL) {
                   dprint("avioInstInit2: ERROR - can't create message queue %s\n", mq_name);
                 }
               
                 if(vxbIntConnect(pInst, pIntRes, (VOIDFUNCPTR)intProcessCommercial, pInst) != OK) {
                   dprint("avioInstInit2: ERROR - can't connect ISR\n", 0);
                 }
                 else {
#ifdef SPIN_LOCK_PROTECT
                   spinLockIsrInit(&spinlock[deviceIndex],0);
#endif
                   if(vxbIntEnable(pInst, pIntRes) != OK) {
                     dprint("avioInstInit2: ERROR - can't enable interrupt\n", 0);
                   }
                 }
                 break;
               /* Setup ISR for PCI Express devices (RAR-XMC) */
               case AVIO_PCI_RARXMC_DEVID:
               case AVIO_PCI_RAR_MPCIE_DEVID:
                 /* create a Commercial Product Message Queue for this device for interrupt notification purposes */
                 sprintf(mq_name,"/%s%d", AVIO_COM_MQ_NAME, deviceIndex);
                 pDrvCtrl->msg_q_com_id = msgQOpen(mq_name, 100, 4, MSG_Q_FIFO, OM_CREATE, NULL);
                 if(pDrvCtrl->msg_q_com_id == MSG_Q_ID_NULL) {
                   dprint("avioInstInit2: ERROR - can't create message queue %s\n", mq_name);
                 }
                 else {
                   if(vxbIntConnect(pInst, pIntRes, (VOIDFUNCPTR)intProcessCommercialPcie, pInst) != OK) {
                     dprint("avioInstInit2: ERROR - can't connect ISR\n", 0);
                   }
#ifdef SPIN_LOCK_PROTECT
                   spinLockIsrInit (&spinlock[deviceIndex],0);
#endif
                   if(vxbIntEnable (pInst, pIntRes) != OK) {
                     dprint("avioInstInit2: ERROR - can't enable interrupt\n", 0);
                   }
                 }
                 break;
               /* Interrupts are not supported with the P708/RP-708 or P-DIS */
               case AVIO_PCI_RP708_DEVID:
               case AVIO_PCI_P708_DEVID:
               case AVIO_PCI_PDIS_DEVID:
                 break;
             }
             if(pIntRes != NULL) {
               pResIrq = (VXB_RESOURCE_IRQ *)pIntRes->pRes;
               if(pResIrq != NULL) {
                 pDrvCtrl->irq = pResIrq->lVec;
               }
             }
             pDrvCtrl->devcnt++;
#ifdef VXW_DRIVER_OPTION
             /* for RTP support: set the AVIO device */
             if(avioIoDevRegister(pInst) != OK)
               errnoSet(EFAULT);
#endif               
             return OK;
          }
          else {
             break;
          }
       }
       else {
         break;
       }
     }
     else {
       break;
     }
   }

   /* disconnect from VxBus */
   dprint("avioInstInit2: This should not happen\n",0);
   
   vxAtomicDec(&avioRefCnt);
   avioVxbResourceFree (pInst, pDrvCtrl);
   free(pDrvCtrl);
   pDrvCtrl = NULL;
   vxbDevSoftcSet (pInst, pDrvCtrl);
   errnoSet(EFAULT);
   
   return ERROR;
}

/*********************************************************************************************************************
** NAME:
** intProcess()
**
** DESCRIPTION:
** ISR for legacy 1553 boards running UCA f/w v4/5.
**
** ARGUMENTS:
** VXB_DEV_ID pInst
**
** RETURNS
**    N/A
**********************************************************************************************************************/
void intProcess(VXB_DEV_ID pInst)
{
   unsigned short control_data=0;  /* control register contents */
   unsigned short * chanHiPtr[4];  /* pointers to each 1553 channel host interface */
   UINT int_bit=0;
   int chan=0;
   MSG_Q_ID mq_id;
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl=NULL;

#ifdef SPIN_LOCK_PROTECT
   UINT32 deviceIndex = vxbDevUnitGet (pInst);
#endif
   /* setup the interrupt bit mask */
   int_bit = UCA_INT_ASSERTED;
   flipw(&int_bit);
   pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)vxbDevSoftcGet (pInst);

   /* define the pointers to each 1553 channel's control register offset into the device host interface */
   chanHiPtr[0] = (unsigned short *)(pDrvCtrl->baseAddressBar0 + UCA16_1553_CTL_REG);
   chanHiPtr[1] = (unsigned short *)(pDrvCtrl->baseAddressBar0 + UCA16_1553_CH1_OFF + UCA16_1553_CTL_REG);
   chanHiPtr[2] = (unsigned short *)(pDrvCtrl->baseAddressBar0 + UCA16_1553_CH2_OFF + UCA16_1553_CTL_REG);
   chanHiPtr[3] = (unsigned short *)(pDrvCtrl->baseAddressBar0 + UCA16_1553_CH3_OFF + UCA16_1553_CTL_REG);

   mq_id = pDrvCtrl->msg_q_id;

   /* Clear the interrupt */
   /* read the CSC control reg then acknowledge via the int ack reg */
#ifdef SPIN_LOCK_PROTECT
   spinLockIsrTake(&spinlock[deviceIndex]);
#endif
   /* check each potential channel location */
   for(chan = 0; (chan < pDrvCtrl->dev_chan) && (chan < MAX_CHANNELS); chan++) {
     control_data = *chanHiPtr[chan];
     /* if the "interrupt pending" bit is high for this channel, acknowledge the interrupt */
     if(control_data & int_bit) {
       /* write 0 to the write_interrupt_bit */
       (chanHiPtr[chan])[UCA16_1553_ACK_REG] = 0;

       /* Send a message to the host thread indicating the respective 1553 channel interrupt event occurred. */
       msgQSend (mq_id, (char *)&mq_data[chan], sizeof(unsigned), NO_WAIT, MSG_PRI_NORMAL);
     }
   }

#ifdef SPIN_LOCK_PROTECT
   spinLockIsrGive(&spinlock[deviceIndex]);
#endif
}

/*********************************************************************************************************************
** NAME:
** intProcess32()
**
** DESCRIPTION:
** ISR for Avionics IO 1553 and MultiProtocol boards (UCA32 f/w v6).
**
** ARGUMENTS:
** VXB_DEV_ID pInst
**
** RETURNS
**    N/A
**********************************************************************************************************************/
void intProcess32(VXB_DEV_ID pInst)
{
   UINT control_data=0;                       /* control register contents */
   UINT temp=0;                               /* interrupt acknowledge data word */
   unsigned int ctrlRegOffset=0;              /* offset into the host interface for the next channel */
   int chan=0;                                /* 1553 channel index */
   MSG_Q_ID mq_id;                            /* local 1553 interrupt event message queue handle, shared with API */
   unsigned int * interruptAckRegPtr;         /* pointer to the device host interface */
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl = NULL;    /* local reference to this device's Driver Control structure */
#ifdef SPIN_LOCK_PROTECT
   UINT32 deviceIndex=vxbDevUnitGet (pInst);  /* device index */
#endif

   /* access this device's Driver Control structure and 1553 interrupt event message queue handle */
   pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)vxbDevSoftcGet (pInst);
   mq_id = pDrvCtrl->msg_q_id;

   /* assign the local reference to this device's base host interface */
   interruptAckRegPtr = (unsigned int *)pDrvCtrl->baseAddressBar0;

#ifdef SPIN_LOCK_PROTECT
   spinLockIsrTake(&spinlock[deviceIndex]);
#endif
   /* check each installed 1553 channel to determine if one or more of them asserted the interrupt */
   ctrlRegOffset = UCA32_1553_CTL_REG;
   for(chan = 0; (chan < pDrvCtrl->dev_chan) && (chan < MAX_CHANNELS); chan++) {
     control_data = *(&interruptAckRegPtr[ctrlRegOffset]);
     fliplong(&control_data);
     if(control_data & UCA_INT_ASSERTED) {
       /* Clear the interrupt in the respective channel's 1553 Control Register */
       temp = UCA32_CLEAR_INT;
       fliplong(&temp);
       *(&interruptAckRegPtr[ctrlRegOffset]) = temp;

       /* Send a message to the 1553-specific message queue, indicating to the host ISR thread an interrupt event occurred. */
       msgQSend(mq_id, (char *)&mq_data[chan], sizeof(unsigned), NO_WAIT, MSG_PRI_NORMAL);
     }

     /* increment for the next channel */
     ctrlRegOffset += UCA32_1553_CH_OFST;
   }

   /* if this device contains an ARINC component (I.E. a combo card), check for an ARINC firmware interrupt assertion */
   if(pDrvCtrl->arinc) {
     control_data = *(&interruptAckRegPtr[MULT_PT_ARINC_OFST]);
     fliplong(&control_data);
     /* if an ARINC interrupt is asserted, handle it */
     if(control_data & ARINC_INT_ASSERTED) {
       /* Clear the interrupt in the ARINC Global Enable Register */
       temp = ARINC_CLEAR_INT;
       fliplong(&temp);
       *(&interruptAckRegPtr[MULT_PT_ARINC_OFST]) = temp;

       /* Send a message to the "commercial product" message queue, indicating to the host ISR thread an interrupt event occurred. */
       msgQSend(pDrvCtrl->msg_q_com_id, (char *)&mq_data[0], sizeof(unsigned), NO_WAIT, MSG_PRI_NORMAL);
     }
   }

#ifdef SPIN_LOCK_PROTECT
   spinLockIsrGive(&spinlock[deviceIndex]);
#endif
}

/*********************************************************************************************************************
** NAME:
** intProcessCommercial()
**
** DESCRIPTION:
** ISR for PCI based Commercial Avionics IO cards (x30 and P-SER), all having a PLX PCI9056 interface
**
** ARGUMENTS:
** VXB_DEV_ID pInst
**
** RETURNS
**    N/A
**********************************************************************************************************************/
void intProcessCommercial(VXB_DEV_ID pInst)
{
   unsigned int control_data=0;                     /* control register contents */
   unsigned int temp=0;                             /* interrupt acknowledge data word */
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl=NULL;            /* local reference to this device's Driver Control structure */
#ifdef SPIN_LOCK_PROTECT
   UINT32 deviceIndex = vxbDevUnitGet (pInst);             /* device index */
#endif
   
   /* access this device's Driver Control structure */
   pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)vxbDevSoftcGet (pInst);

#ifdef SPIN_LOCK_PROTECT
   spinLockIsrTake(&spinlock[deviceIndex]);
#endif
   /* assign the local reference to this device's base host interface */
   control_data = *(unsigned int *)pDrvCtrl->baseAddressBar2;
   temp = control_data;
   fliplong(&control_data);
   /* if a "commercial product" interrupt is asserted, handle it */
   if(control_data & ARINC_INT_ASSERTED) {
     /* Clear the interrupt */
     *(unsigned int *)pDrvCtrl->baseAddressBar2 = temp; /* Clear the interrupt assertion by writing control register value back */

     /* Send a message to the "commercial product" message queue, indicating to the host ISR thread an interrupt event occurred. */
     msgQSend(pDrvCtrl->msg_q_com_id, (char *)&mq_data[0], sizeof(unsigned), NO_WAIT, MSG_PRI_NORMAL);
   }
#ifdef SPIN_LOCK_PROTECT
   spinLockIsrGive(&spinlock[deviceIndex]);
#endif
}


/*********************************************************************************************************************
** NAME:
** intProcessCommercialPcie()
**
** DESCRIPTION:
** ISR for PCI Express Commercial Avionics IO cards (RAR-XMC)
**
** ARGUMENTS:
** VXB_DEV_ID pInst
**
** RETURNS
**    N/A
**********************************************************************************************************************/
void intProcessCommercialPcie(VXB_DEV_ID pInst)
{
   unsigned int control_data=0;                     /* control register contents */
   unsigned int temp=0;                             /* interrupt acknowledge data word */
   unsigned int * interruptAckRegPtr=NULL;          /* pointer to the device host interface */
   AVIO_DEVICE_CTRL_TYPE *pDrvCtrl = NULL;          /* local reference to this device's Driver Control structure */
#ifdef SPIN_LOCK_PROTECT
   UINT32 deviceIndex = vxbDevUnitGet (pInst);      /* device index */
#endif 
   
   /* access this device's Driver Control structure */
   pDrvCtrl = (AVIO_DEVICE_CTRL_TYPE *)vxbDevSoftcGet (pInst);

   /* assign the local reference to this device's base host interface */
   interruptAckRegPtr = (unsigned int *)pDrvCtrl->baseAddressBar0;

#ifdef SPIN_LOCK_PROTECT
   spinLockIsrTake(&spinlock[deviceIndex]);
#endif
   /* assign the local reference to this device's base host interface */
   control_data = *(&interruptAckRegPtr[0]);
   fliplong(&control_data);
   /* if a "commercial product" interrupt is asserted, handle it */
   if(control_data & ARINC_INT_ASSERTED) {
     /* Clear the interrupt */
     temp = ARINC_CLEAR_INT;
     fliplong (&temp);
     *(&interruptAckRegPtr[0]) = temp; /* Atomic write to "clear the interrupt asserted bit" to clear the interrupt */

     /* Send a message to the "commercial product" message queue, indicating to the host ISR thread an interrupt event occurred. */
     msgQSend (pDrvCtrl->msg_q_com_id, (char *)&mq_data[0], sizeof(unsigned), NO_WAIT, MSG_PRI_NORMAL);
   }
#ifdef SPIN_LOCK_PROTECT
   spinLockIsrGive(&spinlock[deviceIndex]);
#endif
}

/*********************************************************************************************************************
** NAME:
** vxbDrvUnlink()
**
** DESCRIPTION:
** Calls vxbDrvUnlink for all instances of the driver.
**
** ARGUMENTS:
** VXB_DRV_CONTROL * pDrvCtrl
**
** RETURNS
**    OK or ERROR
**
**********************************************************************************************************************/
STATUS avioVxbDrvUnlink(VXB_DRV_CONTROL * pDrvCtrl)
{
   int i;
   STATUS res=OK;
   

   for(i = 0; i < DRV_MAX_DEVICES; i++) {
     if(avioInst[i]) {
       if((VXB_DRV_UNLINK(avioInst[i], pDrvCtrl)) == ERROR) {
         res = ERROR;
       }
     }
   }
   
   return (res);
}

/*********************************************************************************************************************
** NAME:
** driverControl()
**
** DESCRIPTION:
** Calls driverControl for all instances of the driver.
**
** ARGUMENTS:
** VXB_DRV_CONTROL * pDrvCtrl
**
** RETURNS
**    OK or ERROR
**
**********************************************************************************************************************/
STATUS driverControl(VXB_DRV_CONTROL * pDrvCtrl)
{
   int i;
   STATUS res=OK;

   for(i = 0; i < DRV_MAX_DEVICES; i++) {
     if(avioInst[i]) {
       if((DRIVER_CONTROL (avioInst[i], pDrvCtrl)) == ERROR) {
         res = ERROR;
       }
     }
   }

   return (res);
}

/*********************************************************************************************************************
**
** NAME:
** avioVxbResourceFree()
**
** DESCRIPTION:
** Frees VxBus2 memory and interrupt resources allocated for this device instance.
**
** ARGUMENTS:
** pInst    - VxBus device instance.
** pDrvCtrl - device instance data
**
** RETURN:
** OK or ERROR
**
**********************************************************************************************************************/
LOCAL STATUS avioVxbResourceFree(VXB_DEV_ID pInst, AVIO_DEVICE_CTRL_TYPE *pDrvCtrl)
{
   if((pInst == NULL) || (pDrvCtrl == NULL)) {
     return ERROR;
   }

   if(pDrvCtrl->pIntRes) {
     vxbResourceFree (pInst, pDrvCtrl->pIntRes);
     pDrvCtrl->pIntRes = NULL;
     pDrvCtrl->irq = 0;
   }

   if(pDrvCtrl->pMemRes0) {
     vxbResourceFree (pInst, pDrvCtrl->pMemRes0);
     pDrvCtrl->pMemRes0 = NULL;
     pDrvCtrl->baseAddressBar0 = NULL;
     pDrvCtrl->membase = 0UL;
     pDrvCtrl->baseHandle[0] = NULL;
   }

   if(pDrvCtrl->pMemRes2) {
     vxbResourceFree (pInst, pDrvCtrl->pMemRes2);
     pDrvCtrl->pMemRes2 = NULL;
     pDrvCtrl->baseAddressBar2 = NULL;
     pDrvCtrl->membaseBar2 = 0UL;
     pDrvCtrl->baseHandle[2] = NULL;
   }

   return OK;
}

/*****************************************************************************
* 
* vxbDevMethodRun - run the specified method if supported for the driver 
* 
* This is a bridge between VxBus1 and VxBus2, allowing customer code to run
* without modification.  It allows execution of the following methods for
* the AVIO driver: driverControl and vxbDrvUnlink.
* 
* RETURNS
* Results of running the method, if it exists.
* ERROR otherwise.
*/
STATUS vxbDevMethodRun(VXB_METHOD_ID method, void *pArg)
{
   int i;
   VXB_METHOD_ID  m;
   char * drvName = AVIO_DRV_NAME;
   
   /* parameter validation */
   if((method == 0) || (pArg == NULL)) {
     return ERROR;
   }
   
   /* check if this is our driver */
   if((strncmp(drvName, ((VXB_DRV_CONTROL *)pArg)->driverName, strlen (drvName))) != 0) {
     return ERROR;
   }
   
   /* scan the local table for a match */
   for(i = 0, m = avioVxbMethods[i].drvMethodId; m != 0; ) {
     if(m == method) {
       return (avioVxbMethods[i].handler ((VXB_DRV_CONTROL *)pArg));
     }
     i++;
     m = avioVxbMethods[i].drvMethodId;
   }
   
   /* should not get here, unless we don't have a matching method */
   return ERROR;
}
